from typing import List
import numpy as np
import pandas as pd
from typeguard import typechecked

from neurokin.utils.features_extraction.core import FeatureExtraction


class Height(FeatureExtraction):
    """
    Computes Height of selected markers and one selected coordinate (x, y, z). By default it sets the lowest point to 0
    so all other points are referenced to it.
    """
    extraction_target = "misc"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: List, coord: str, normalize_min: bool = True, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        df_feat_list = []
        df_feat = pd.DataFrame()
        target_markers = [coord for marker in target_bodyparts for coord in bodyparts_coordinates if
                                 marker in coord]
        target_markers_coords = [marker for marker in target_markers if coord in marker]

        for bodypart in target_markers_coords:
            feat = df[bodypart].values
            if normalize_min:
                feat = feat - np.nanmin(feat)
            df_feat_list.append(pd.DataFrame(feat, columns=[f"{bodypart}_height"]))
        if df_feat_list:
            df_feat = pd.concat(df_feat_list, axis=1)
        return df_feat



class FwdMovement(FeatureExtraction):
    """
    Computes Forward Movement of selected markers and one selected coordinate (x, y, z), by returning the first
    derivative. By default it sets the lowest point to 0 so all other points are referenced to it.
    """
    extraction_target = "misc"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: List, coord: str, normalize_min: bool = True, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        df_feat_list = []
        df_feat = pd.DataFrame()
        target_markers = [coord for marker in target_bodyparts for coord in bodyparts_coordinates if
                                 marker in coord]
        target_markers_coords = [marker for marker in target_markers if coord in marker]

        for bodypart in target_markers_coords:
            feat = df[bodypart].values
            if normalize_min:
                feat = feat - np.nanmin(feat)
            feat = np.diff(feat)
            df_feat_list.append(pd.DataFrame(feat, columns=[f"{bodypart}_fwd_movement"]))

        if df_feat_list:
            df_feat = pd.concat(df_feat_list, axis=1)
        return df_feat


import pickle as pkl

import numpy as np
import pandas as pd


def load_dataset(dataset_path):
    """
    Load existing dataset dictionary, avoiding waiting time for processing

    :param dataset_path: path including filename to the dataset. Expects pickle files
    :return:
    """
    return pkl.load(open(dataset_path, "rb"))


def save_data(data, filename):
    """
    Saves data to pickle files, if the data is not empty. Else returns a ValueError.
    :param data:
    :param filename: filename with path.
    :return:
    """
    if data is None:
        raise ValueError("The dataset has not been initialized")
    with open(f"{filename}.pkl", "wb") as handle:
        pkl.dump(data, handle)


def condense_distribution_event_types(percentage_states_df):
    """
    Condenses from the dataset five categories analysis
    ["fog_active", "fog_rest", "nlm_active", "nlm_rest", gait] to only
    three categories ["fog", "nlm", "gait"]

    :param cond_animal_event_dict: states dictionary structured as condition, animal, events.
    :return:  same dictionary with 3 conditions instead of 5
    """
    five2three = percentage_states_df.copy()
    five2three["event_fog"] = five2three["event_fog_active"] + five2three["event_fog_rest"]
    five2three["event_nlm"] = five2three["event_nlm_active"] + five2three["event_nlm_rest"]
    five2three.drop(["event_fog_active", "event_nlm_active", "event_fog_rest", "event_nlm_rest"], inplace=True, axis=1)
    return five2three


def mean_psds(psds_list):
    """
    Returns mean psds if the list is full

    :param psds_list: list of power spectral density arrays
    :return: mean on axis 0
    """
    if psds_list:
        return np.mean(psds_list, axis=0)


def get_per_animal_psds_df(psds_dataset, condense=False):
    """
    Given the overall psds dataset, computes the average values, per state, per animal

    :param psds_dataset: overall dataset
    :param condense: wheter to condense to the 3 states format or keep the 5 format
    :return: grouped-by-animal dataset
    """
    psds_means = psds_dataset.copy()
    events_col = [c for c in psds_means.columns if c.startswith("event")]
    psds_means.drop(["date", "run"], inplace=True, axis=1)
    psds_means = psds_means.groupby(["subject", "condition"], as_index=False)[events_col].sum(min_count=1)
    psds_means = psds_means.applymap(lambda x: [] if x is None else x)

    if condense:
        psds_means["event_nlm"] = psds_means["event_nlm_rest"] + psds_means["event_nlm_active"]
        psds_means["event_fog"] = psds_means["event_fog_rest"] + psds_means["event_fog_active"]
        psds_means.drop(["event_nlm_rest", "event_fog_rest", "event_nlm_active", "event_fog_active"],
                        inplace=True, axis=1)

    events_col = [c for c in psds_means.columns if c.startswith("event")]
    psds_means[events_col] = psds_means[events_col].applymap(mean_psds, na_action="ignore")
    return psds_means


def get_group_split(test_sbj_list, df):
    """
    Splits the dataset in two groups depending on which subjects are in the test group and which not (control)

    :param test_sbj_list: list of subject IDs that belong to teh test group
    :param animals_avg_dataset: dictionary containing the psds organized by condition, animal and event
    :return: split dictionary between test_group and sham_group
    """
    subject_groups = df.copy()
    subject_groups["group"] = subject_groups["subject"].apply(lambda x: x in test_sbj_list)
    return subject_groups


def compute_duration(timestamps_lists):
    time = 0
    for times in timestamps_lists:
        time += times[1] - times[0]
    return time


def compute_percentage(array):
    return 100 * array / np.sum(array)


def compute_events_percentage(events_dataset):
    """
    Computes the average percentage of time spent in each state for each animal. It normalizes every state duration to
    the total length of the run, then takes the average across all the runs available for that condition.

    :param events_dataset: dataset with event timestamps
    :return: dataframe containing condition, animal and event containing the average time spent in each state
    """
    durations = events_dataset.copy()
    events_col = [col for col in durations.columns if col.startswith("event")]
    durations[events_col] = durations[events_col].applymap(compute_duration).apply(compute_percentage, axis=1)
    durations.drop("date", inplace=True, axis=1)
    means = durations.groupby(["condition", "subject"], as_index=False).mean(numeric_only=True)
    return means


def compute_ci(array):
    """
    Computes the confidence interval given hardcoded value of 1.96.

    :param array: array to compute the confidence interval on
    :return: confidence interval value
    """
    ci = 1.96 * np.std(array) / np.sqrt(len(array))
    return ci


def get_state_graph_stats(group_cond_df, stat="std"):
    """
    Takes the group dataset and returns a dataframe with mean, upper bound and lower bound to be used to graph.

    :param group_cond_df:
    :param stat: which statistic to use to compute the upper and lowe bounds. Takes "std" for standard deviation,
                 "sem" for standard error of mean and "ci" for confidence interval.
    :return: df
    """

    mean_df = group_cond_df.groupby(["group", "condition"], as_index=False).mean(numeric_only=True)
    if stat == "std":
        dist_df = group_cond_df.groupby(["group", "condition"], as_index=False).std(numeric_only=True)
    elif stat == "ci":
        dist_df = group_cond_df.groupby(["group", "condition"], as_index=False).apply(compute_ci, axis=0)
    elif stat == "sem":
        dist_df = group_cond_df.groupby(["group", "condition"], as_index=False).sem(numeric_only=True)
    else:
        raise ValueError(f"The statistic value {stat} is unsupported, please choose between std, sem or ci")

    mean_df = pd.melt(mean_df, id_vars=['group', 'condition'], value_vars=['event_gait', 'event_fog', 'event_nlm'],
                      var_name='event_type', value_name='mean')

    dist_df = pd.melt(dist_df, id_vars=['group', 'condition'], value_vars=['event_gait', 'event_fog', 'event_nlm'],
                      var_name='event_type', value_name='distribution')
    stats = mean_df.copy()
    stats["lower_bound"] = mean_df["mean"] - dist_df["distribution"]
    stats["upper_bound"] = mean_df["mean"] + dist_df["distribution"]

    return stats


def get_runs_list(experiment_structure, skip_subjects, skip_conditions):
    """
    Retrieves all the runs unique strings identifying day, subject, condition and run number

    :param experiment_structure: dictionary containing the experiment structure
    :param skip_subjects: which subjects to skip
    :param skip_conditions: which condition to skip
    :return: flat list with all the unique strings identifying runs
    """
    experiment_combos = []
    dates = [date for date in experiment_structure.keys()]
    for date in dates:
        subjects = [subject for subject in experiment_structure[date].keys() if subject not in skip_subjects]
        for subject in subjects:
            conditions = [condition for condition in experiment_structure[date][subject].keys() if
                          condition not in skip_conditions]
            for condition in conditions:
                runs = [str(run) for run in experiment_structure[date][subject][condition]]
                runs = ["0" + run if len(run) == 1 else run for run in runs]
                for run in runs:
                    experiment_combos.append([date, subject, condition, run])

    return experiment_combos



import numpy as np
from neurokin.utils.experiments.neural_states_helper import compute_ci


def plot_psd_single_state(ax, df, group, condition, state, freqs, color, idx_min, idx_max, stat="std", linewidth=3):
    """
    Plots the Power Spectral Denisty trace of a single state, on a given ax.

    :param ax: axis to plot on
    :param group: which subject group to plot
    :param condition: which experimental condition to plot
    :param state: which state to plot
    :param freqs: frequencies corresponding to the full array of PSD values
    :param idx_min: index of the minimum frequency to plot
    :param idx_max: index of the maximum frequency to plot
    :param color: color of the trace
    :param stat: which stat to use to compute the shading, default standard deviation
    :param linewidth: how thick to draw the mean line.
    :return:
    """

    f = freqs[idx_min:idx_max]
    data = df[(df["group"] == group) & (df["condition"] == condition)][state]
    data_mean = data.mean(axis=0)[idx_min:idx_max]
    ax.plot(f, data_mean, color=color, linewidth=linewidth)
    add_shades(ax, f, data, idx_min, idx_max, color, stat=stat)


def add_shades(ax, f, df, idx_min, idx_max, color, stat="std"):
    """
    Helper function to plot shades.

    :param ax: the ax to plot on
    :param f: frequencies to use on the x-axis
    :param df: the data to be plot (array like of states)
    :param idx_min: minimum frequency index to plot
    :param idx_max: maximum frequency index to plot
    :param color: color to plot the shade in, with 0.3 alpha factor
    :param stat: which statistics to use for shades. "std" for standard deviation,
                "sem" for standard error of mean, "ci" for confidence interval
    :return:
    """
    data = df.dropna().values
    if stat == "sem":
        s = data.sem(axis=0)
        s = data.std(axis=0, ddof=1) / np.sqrt(np.size(data))
    elif stat == "ci":
        s = data.apply(compute_ci, axis=0)
        s = compute_ci(data)
    elif stat == "std":
        s = data.std(axis=0)
    else:
        raise ValueError(f"The statistic value {stat} is unsupported, please choose between std, sem or ci")

    lower = data.mean(axis=0) - s
    upper = data.mean(axis=0) + s
    ax.fill_between(f, lower[idx_min:idx_max], upper[idx_min:idx_max], alpha=0.3, color=color)



from typing import Dict
import pandas as pd
from typeguard import typechecked

from neurokin.utils.features_extraction.commons import (compute_angle,
                                                        compute_angle_correlation,
                                                        compute_angle_acceleration,
                                                        compute_angle_velocity,
                                                        compute_angle_phase)
from neurokin.utils.features_extraction.core import FeatureExtraction


class Angle(FeatureExtraction):
    extraction_target = "joints"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: Dict, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        df_feat_list = []
        df_feat = pd.DataFrame()
        for joint, bodyparts in target_bodyparts.items():
            target_markers_coords = [coord for marker in bodyparts for coord in bodyparts_coordinates if
                                     marker in coord]
            feat = compute_angle(df[target_markers_coords].values)
            df_feat_list.append(pd.DataFrame(feat, columns=[f"{joint}_angle"]))

        if df_feat_list:
            df_feat = pd.concat(df_feat_list, axis=1)
        return df_feat


class AngleVelocity(FeatureExtraction):
    extraction_target = "joints"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: Dict, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        df_feat_list = []
        df_feat = pd.DataFrame()
        for joint, bodyparts in target_bodyparts.items():
            target_markers_coords = [coord for marker in bodyparts for coord in bodyparts_coordinates if
                                     marker in coord]
            feat = compute_angle_velocity(df[target_markers_coords].values)
            df_feat_list.append(pd.DataFrame(feat, columns=[f"{joint}_angle_velocity"]))

        if df_feat_list:
            df_feat = pd.concat(df_feat_list, axis=1)
        return df_feat


class AngleAcceleration(FeatureExtraction):
    extraction_target = "joints"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: Dict, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        df_feat_list = []
        df_feat = pd.DataFrame()
        for joint, bodyparts in target_bodyparts.items():
            target_markers_coords = [coord for marker in bodyparts for coord in bodyparts_coordinates if
                                     marker in coord]
            feat = compute_angle_acceleration(df[target_markers_coords].values)
            df_feat_list.append(pd.DataFrame(feat, columns=[f"{joint}_angle_acceleration"]))

        if df_feat_list:
            df_feat = pd.concat(df_feat_list, axis=1)
        return df_feat


class AngleCorrelation(FeatureExtraction):
    extraction_target = "joints"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: Dict, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        df_feat_list = []
        df_feat = pd.DataFrame()
        for joint, bodyparts in target_bodyparts.items():
            target_markers_coords = [coord for marker in bodyparts for coord in bodyparts_coordinates if
                                     marker in coord]
            feat = compute_angle_correlation(df[target_markers_coords].values)
            df_feat_list.append(pd.DataFrame(feat, columns=[f"{joint}_angle_correlation"]))

        if df_feat_list:
            df_feat = pd.concat(df_feat_list, axis=1)
        return df_feat


class AnglePhase(FeatureExtraction):
    extraction_target = "joints"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: Dict, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        df_feat_list = []
        df_feat = pd.DataFrame()
        for joint, bodyparts in target_bodyparts.items():
            target_markers_coords = [coord for marker in bodyparts for coord in bodyparts_coordinates if
                                     marker in coord]
            feat = compute_angle_phase(df[target_markers_coords].values)
            df_feat_list.append(pd.DataFrame(feat, columns=[f"{joint}_angle_phase"]))

        if df_feat_list:
            df_feat = pd.concat(df_feat_list, axis=1)
        return df_feat


class CustomJointFeatures(FeatureExtraction):
    extraction_target = "joints"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: Dict, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        df_feat_list = []
        df_feat = pd.DataFrame()
        try:
            feature_names = kwargs.get("feature_names")
        except KeyError:
            raise KeyError(f"No name found for the Custom Join Function. "
                           f"Please add valid feature_names in the config file.")
        for name in feature_names:
            try:
                func = kwargs.get("custom_features")[name]
            except KeyError:
                raise KeyError(f"No function found with name: {name}, please provide a valid name and function "
                               f"name when calling extract_features")
            for joint, bodyparts in target_bodyparts.items():
                target_markers_coords = [coord for marker in bodyparts for coord in bodyparts_coordinates if
                                         marker in coord]
                feat = func(df[target_markers_coords].values)
                df_feat_list.append(pd.DataFrame(feat, columns=[f"{joint}_{name}"]))

        if df_feat_list:
            df_feat = pd.concat(df_feat_list, axis=1)
        return df_feat



import numpy as np
import pandas as pd
from scipy import signal


#TESTME with mock df
def tilt_correct(df, reference_marker, columns_to_correct):
    """
    If the runway is not perfectly aligned there can be a linear trend in one of the axis.
    This function computes the linear trend from a reference marker and applies it to all the columns passed.
    E.g. use left mtp z axis to compute the trend, then subtract the trend from all columns representing the z axis
    of a marker.

    """

    trend = signal.detrend(df[reference_marker]) - df[reference_marker]
    df_tilt_corrected = df.apply(lambda x: x.add(trend, axis=0) if x.name in columns_to_correct else x)
    return df_tilt_corrected

#TESTME with mock df
def shift_correct(df, reference_marker, columns_to_correct):
    #TODO compliance with DLC?
    """
    If the origin is not set to the beginning of the runway (e.g. set to the center) one of the axis will have negative
    values. This functions shifts all the columns to be corrected by the minimum value of the reference marker.
    The reference marker should be the one farther in the back.

    """

    shift = abs(min(df[reference_marker])) if min(df[reference_marker]) < 0 else 0
    df_shift_corrected = df.apply(lambda x: x.add(shift, axis=0) if x.name in columns_to_correct else x)
    return df_shift_corrected


def check_correct_columns_extraction(actual, expected, side):
    if actual == expected:
        return
    else:
        raise ValueError("Warning: the number of selected columns for the " + side + " side [" + str(actual) +
                         "] does not match the \n"
                         "expected " + str(expected) +
                         ". Please check if there are ambiguity in the column names.")


def get_unilateral_df(df, side="", name_starts_with=False, name_ends_with=False,
                      column_names=None, expected_columns_number=None):
    if column_names is None:
        column_names = []
    if name_starts_with:
        df_side = df.loc[:, df.columns.str.startswith(side)]
    elif name_ends_with:
        df_side = df.loc[:, df.columns.str.endswith(side)]
    elif column_names:
        df_side = df.loc[:, column_names]
    else:
        print("WARNING: no columns selected for side " + side + ". Please check if this is expected behaviour. \n"
                                                                "If not, you should set the side and either starts "
                                                                "or ends with, alternatively pass column_names.")
        return
    if expected_columns_number:
        check_correct_columns_extraction(len(df_side.columns.tolist()), expected_columns_number, side)

    return df_side



def create_empty_features_df(self, bodyparts, features):
    # TODO check shape passed: not matching
    dataFrame = None
    steps_number = max([len(self.right_mtp_land), len(self.left_mtp_land)])
    a = np.full((steps_number), np.nan)
    for bodypart in bodyparts:
        pdindex = pd.MultiIndex.from_product(
            [features, [bodypart]],
            names=["feature", "bodypart"])
        frame = pd.DataFrame(a, columns=pdindex, index=range(0, steps_number))
        dataFrame = pd.concat([frame, dataFrame], axis=1)
    return dataFrame


from setuptools import setup, find_packages

setup(
    name='neurokin',
    version='0.0.1',
    description='package to support neural and data analysis',
    url='https://github.com/WengerLab/neurokin',
    author='Elisa L. Garulli',
    author_email='e.garulli@charite.de',
    license='BSD 2-clause',
    packages=find_packages(),
    install_requires=['c3d==0.5.2',
                      'fooof==1.0.0',
                      'tdt==0.5.3',
                      'numpy==1.24.4',
                      'pandas==2.1.4',
                      'scipy==1.11.4',
                      'matplotlib==3.7.3',
                      'typeguard==4.4.2',
                      'pyyaml==6.0.2'
                      ],
extras_require={
        "dev": ["pytest"],
        "docs": ["sphinx", "nbsphinx", "sphinxawesome-theme", "pygments"]
    },

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)